/* 
 * 
 * eulerForce v0.9.0
 * 
 * 2013-11-25
 *
 * 
 * 
 * eulerForce -- Laying Out Euler Diagrams Using a Force-Directed Approach	    
 * 		http://www.eulerdiagrams.org/eulerForce
 * 
 * 
 * 		Copyright (C) 2013, Luana Micallef and Peter Rodgers. 
 * 		All rights reserved.
 * 		
 * 
 * 		This file is part of eulerForce.
 * 			
 * 		eulerForce is free software: you can redistribute it and/or modify
 * 		it under the terms of the GNU General Public License as published 
 * 		by the Free Software Foundation, either version 3 of the License, 
 * 		or (at your option) any later version.
	
 * 		eulerForce is distributed in the hope that it will be useful,
 * 		but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * 		GNU General Public License for more details.
	
 * 		A copy of the GNU General Public License is provided with 
 * 		eulerForce (in a file named 'COPYING'). Alternatively, see 
 * 		<http://www.gnu.org/licenses/gpl.html>.
 * 			
 */



package euler;

import java.awt.*;
import java.util.*;


/**
 * To instantiate, define and handle Euler diagrams whose curves are polygons
 * 
 */



public abstract class ConcreteDiagram {
	
	protected boolean optimizeContourAngles;
	protected boolean optimizeMeetingPoints = false;
	protected boolean fitCircles;
	public static final int PARALLEL_FUDGE_FACTOR = 4; // how close in degrees lines have to be to be parallel
	public static final int OUTER_FACE_TRIANGULATION_BOUNDARY = 100;
	public static final int CP_OPTIMIZATION_ITERATIONS = 50;
	/** minimum gap between separate contours through TEs, as a percentage of edge length. */
	public static final double CONTOUR_GAP_PERCENT = 0.3;
	public static final int CONCURRENT_OFFSET = 2;
	/** how much gap between concurrent contours*/
	protected int concurrentOffset = CONCURRENT_OFFSET;
	protected ArrayList<ConcreteContour> concreteContours;

	
	
	
	public ConcreteDiagram(ArrayList<ConcreteContour> ccs){
		concreteContours = ccs;		
	}
	
	public ArrayList<ConcreteContour> getConcreteContours() {return concreteContours;}

	
	/**
	 * Scales and centers the contours so that they are maximized
	 * in the given rectangle. This does not change the
	 * aspect ratio of the graph, it is either maximized
	 * in x or y directions.
	 */
	public static void fitContoursInRectangle(ArrayList<ConcreteContour> ccs, int x1, int y1, int x2, int y2) {
		int width = findContoursWidth(ccs);
		int height = findContoursHeight(ccs);
		int rectX1 = x1;
		int rectX2 = x2;
		if(rectX1 > rectX2) {
			rectX1 = x2;
			rectX2 = x1;
		}
		int rectY1 = y1;
		int rectY2 = y2;
		if(rectY1 > rectY2) {
			rectY1 = y2;
			rectY2 = y1;
		}
		int requiredX = rectX2-rectX1;
		int requiredY = rectY2-rectY1;
		
		int centreX = rectX1 + requiredX/2;
		int centreY = rectY1 + requiredY/2;

		double xRatio = requiredX/(double)width;
		double yRatio = requiredY/(double)height;
		// scale to the ratio that will not take the graph over the limits
		double scaleFactor = xRatio;
		if(yRatio < xRatio) {
			scaleFactor = yRatio;
		}
		
		scaleContours(ccs, scaleFactor);
		centreContoursOnPoint(ccs, centreX, centreY);
		
	}
	
	
	/** The width of the contours */
	public static int findContoursWidth(ArrayList<ConcreteContour> ccs) {
		
		int maxX = Integer.MIN_VALUE;
		int minX = Integer.MAX_VALUE;
		
		for(ConcreteContour cc : ccs) {
			Polygon p = cc.getPolygon();
			for(int i = 0; i < p.npoints; i++) {
				int x = p.xpoints[i];
				if(x < minX) {
					minX = x;
				}
				if(x > maxX) {
					maxX = x;
				}
			}
		}

		return maxX-minX;
	}
	
	/** The height of the contours */
	public static int findContoursHeight(ArrayList<ConcreteContour> ccs) {

		int maxY = Integer.MIN_VALUE;
		int minY = Integer.MAX_VALUE;
		
		for(ConcreteContour cc : ccs) {
			Polygon p = cc.getPolygon();
			for(int i = 0; i < p.npoints; i++) {
				int y = p.ypoints[i];
				if(y < minY) {
					minY = y;
				}
				if(y > maxY) {
					maxY = y;
				}
			}
		}
		
		return maxY-minY;
	}
	
	

	/**
	* Finds the centre of the contours, based on forming a rectangle
	* around the limiting nodes and edge bends in the graph.
	*/
	public static Point findContoursCentre(ArrayList<ConcreteContour> ccs) {

		int maxX = Integer.MIN_VALUE;
		int minX = Integer.MAX_VALUE;
		int maxY = Integer.MIN_VALUE;
		int minY = Integer.MAX_VALUE;

		for(ConcreteContour cc : ccs) {
			Polygon p = cc.getPolygon();
			for(int i = 0; i < p.npoints; i++) {
				int x = p.xpoints[i];
				int y = p.ypoints[i];
				if(x > maxX) {
					maxX = x;
				}
				if(x < minX) {
					minX = x;
				}
				if(y > maxY) {
					maxY = y;
				}
				if(y < minY) {
					minY = y;
				}
			}
		}

		int retX = minX + (maxX - minX)/2;
		int retY = minY + (maxY - minY)/2;

		Point ret = new Point(retX,retY);
		return ret;
	}
	/**
	* Finds the centre of the contours, based on forming a rectangle
	* around the limiting nodes and edge bends in the graph.
	*/
	public static Rectangle findContoursBounds(ArrayList<ConcreteContour> ccs) {

		int maxX = Integer.MIN_VALUE;
		int minX = Integer.MAX_VALUE;
		int maxY = Integer.MIN_VALUE;
		int minY = Integer.MAX_VALUE;

		for(ConcreteContour cc : ccs) {
			Polygon p = cc.getPolygon();
			for(int i = 0; i < p.npoints; i++) {
				int x = p.xpoints[i];
				int y = p.ypoints[i];
				if(x > maxX) {
					maxX = x;
				}
				if(x < minX) {
					minX = x;
				}
				if(y > maxY) {
					maxY = y;
				}
				if(y < minY) {
					minY = y;
				}
			}
		}

		Rectangle ret = new Rectangle(minX,minY,maxX-minX,maxY-minY);
		return ret;
	}
	/** Centre the contours on the given point */
	public static void centreContoursOnPoint(ArrayList<ConcreteContour> ccs, int centreX, int centreY) {
		Point centre = findContoursCentre(ccs);
	
		int moveX = centreX - centre.x;
		int moveY = centreY - centre.y;
		
		moveContours(ccs,moveX,moveY);
	}
	/** Scales the contours on the contours centre. */
	public static void scaleContours(ArrayList<ConcreteContour> ccs, double multiplier) {

		if(multiplier == 0.0) {
			return;
		}

		Point centre = findContoursCentre(ccs);

		for(ConcreteContour cc : ccs) {
			Polygon p = cc.getPolygon();
			for(int i = 0; i < p.npoints; i++) {
				p.xpoints[i] = pjr.graph.Util.scaleCoordinate(p.xpoints[i],centre.x,multiplier);
				p.ypoints[i] = pjr.graph.Util.scaleCoordinate(p.ypoints[i],centre.y,multiplier);
			}
			cc.resetArea();
		}
	}	
	/** Move all the points by the values given */
	public static void moveContours(ArrayList<ConcreteContour> ccs, int moveX, int moveY) {
	
		for(ConcreteContour cc : ccs) {
			Polygon p = cc.getPolygon();
			for(int i = 0; i < p.npoints; i++) {
				p.xpoints[i] = p.xpoints[i]+moveX;
				p.ypoints[i] = p.ypoints[i]+moveY;
			}
			cc.resetArea();
		}	
	}

}
