/* 
 * 
 * eulerForce v0.9.0
 * 
 * 2013-11-25
 *
 * 
 * 
 * eulerForce -- Laying Out Euler Diagrams Using a Force-Directed Approach	    
 * 		http://www.eulerdiagrams.org/eulerForce
 * 
 * 
 * 		Copyright (C) 2013, Luana Micallef and Peter Rodgers. 
 * 		All rights reserved.
 * 		
 * 
 * 		This file is part of eulerForce.
 * 			
 * 		eulerForce is free software: you can redistribute it and/or modify
 * 		it under the terms of the GNU General Public License as published 
 * 		by the Free Software Foundation, either version 3 of the License, 
 * 		or (at your option) any later version.
	
 * 		eulerForce is distributed in the hope that it will be useful,
 * 		but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * 		GNU General Public License for more details.
	
 * 		A copy of the GNU General Public License is provided with 
 * 		eulerForce (in a file named 'COPYING'). Alternatively, see 
 * 		<http://www.gnu.org/licenses/gpl.html>.
 * 			
 */



package forcedirected;

import java.awt.Polygon;
import java.awt.geom.Area;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map.Entry;

import euler.ConcreteContour;


/**
 * To handle the zones of an Euler diagram drawn with polygons
 * 
 */


public class Zone {
	
	// Private Data Members
	private String abstractDescription;
	private Area area;
	private ArrayList<Polygon> polygons;
	
	
	// Constructors 
	public Zone (String inAbstractDescription){
		abstractDescription = inAbstractDescription;
		area = new Area();
		polygons = new ArrayList<Polygon>();
	}
	public Zone (String inAbstractDescription, Area inArea, ArrayList<Polygon> inPolygons){
		abstractDescription = inAbstractDescription;
		setArea(inArea);
		setPolygons(inPolygons);
	}
	
	
	// Properties
	public String getAbstractDescription(){
		return abstractDescription;
	}
	
	public void setArea (Area inArea){
		area = inArea;
	}
	public Area getArea(){
		return area;
	}
	
	public void setPolygons(ArrayList<Polygon> inPolygons){
		polygons = inPolygons;
	}
	public ArrayList<Polygon> getPolygons(){
		return polygons;
	}
	

	
	// Instance Methods
	
	//    Checks 
	public boolean isZoneEmpty(){
		return (this.abstractDescription == "");
	}
	
	public boolean containsPoint(Point2D.Double pnt){
		for (Polygon poly : this.polygons){
			if (SimpleDiagramDrawerForceModel.isPointInPolygon(poly, pnt)){
				return true;
			}
		}
		return false;
	}
	
	
	//    Computing Zone Area
	public double computePartialArea(int polygonIndex){
		return (computePolygonArea(polygons.get(polygonIndex)));
	}
	
	public double computeTotalArea(){
		double totalArea = 0;
		
		for (Polygon p : polygons){
			totalArea += computePolygonArea(p);}
		
		return totalArea;
	}
	
	private double computePolygonArea (Polygon p){
		return Utilities.computePolygonArea(p);
	}
	
	
	
	//    Compute Centroid -> returns an average centroid based on the number of polygons it contains
	public Point2D.Double computeCentroid (){
		if (polygons.size()==0){return null;}
		Point2D.Double avgZCentroid = new Point2D.Double(0,0);
		for (Polygon p : polygons){
			Point2D.Double pCentroid = computePolygonCentroid(p);
			avgZCentroid.x += pCentroid.x;
			avgZCentroid.y += pCentroid.y;
		}
		avgZCentroid.x = avgZCentroid.x / polygons.size();
		avgZCentroid.y = avgZCentroid.y / polygons.size();
		
		return avgZCentroid;
	}

	public Point2D.Double computeCentroid (int polygonIndex){
		return computePolygonCentroid((Polygon)polygons.get(polygonIndex));
	}
	public ArrayList<Point2D.Double> computeCentroids (){
		ArrayList<Point2D.Double> centroids = new ArrayList<Point2D.Double>(); 
		for (Polygon p : polygons){
			centroids.add(computePolygonCentroid(p));}
		return centroids;	
	}
	public Point2D.Double computePolygonCentroid (Polygon p){
		return (SimpleDiagramDrawerForceModel.convertToPoint2D_Double(Utilities.findMiddlePointInsidePolygon(p)));
	}
	
	
	
	
	// Class Methods 
	public static ArrayList<Zone> removeEmptyZones (ArrayList<Zone> zones){
		ArrayList<Zone> newZones = (ArrayList<Zone>) zones.clone();
		
		for (Zone z : zones){
			if (z.isZoneEmpty()){
				newZones.remove(z);
			}
		}
		return newZones; 
	}
	
	public static Zone getZoneFromAbstractDescription (String zoneAbstractDescription, ArrayList<Zone> zones){
		
		for (Zone z: zones){
			if (z.getAbstractDescription().equals(zoneAbstractDescription))
				return z;
		}
		return null;
	}
	
	
	
	// More general methods
	
	public static Zone getZoneFromListWithName (ArrayList<Zone> zones, String zoneName){
		for (Zone z : zones){
			if (z.abstractDescription.equals(zoneName)){
				return z;
			}
		}
		return null;
	}
	
	public static ArrayList<Zone> getZones(ArrayList<ConcreteContour> ccs, boolean includeEmptyZones){
		
		HashMap<String, Area> zoneAreas = ConcreteContour.generateZoneAreas(ccs);
		Iterator itr = zoneAreas.entrySet().iterator();
		
		ArrayList<Zone> zones = new ArrayList<Zone>();
		
		while (itr.hasNext()){
			Entry<String,Area> strAreaPair = (Entry<String,Area>)itr.next();
			ArrayList<Polygon> zPolys = (ConcreteContour.polygonsFromArea(strAreaPair.getValue()));
			zones.add(new Zone(strAreaPair.getKey(), strAreaPair.getValue(), zPolys));
		}
		
		if (!includeEmptyZones){
			zones = Zone.removeEmptyZones(zones);
		}
		
		return zones;
	}
}
